"""Contains classes and function for processing Illumina methylation arrays.

It includes methods for extracting methylation information, various
preprocessing techniques, normalization, and data handling.
"""

import collections
import os
import pickle
from functools import reduce
from pathlib import Path

import numpy as np
import pandas as pd
from tqdm import tqdm

from mepylome.dtypes.arrays import ArrayType
from mepylome.dtypes.cache import cache_key, input_args_id, memoize
from mepylome.dtypes.idat import IdatParser
from mepylome.dtypes.manifests import Manifest
from mepylome.dtypes.probes import Channel, ProbeType
from mepylome.utils.varia import MEPYLOME_TMP_DIR, normexp_get_xs

ENDING_GRN = "_Grn.idat"
ENDING_RED = "_Red.idat"
ENDING_GZ = ".gz"
ENDING_SUFFIXES = ("_Grn.idat", "_Red.idat", "_Grn.idat.gz", "_Red.idat.gz")

NEUTRAL_BETA = 0.5


def is_valid_idat_basepath(basepath):
    """Checks if the given basepath(s) point to valid IDAT files."""
    if not isinstance(basepath, list):
        basepath = [basepath]
    basepath = [str(x) for x in basepath]
    return all(
        (
            os.path.exists(x + ENDING_GRN)
            or os.path.exists(x + ENDING_GRN + ENDING_GZ)
        )
        and (
            os.path.exists(x + ENDING_RED)
            or os.path.exists(x + ENDING_RED + ENDING_GZ)
        )
        for x in basepath
    )


def idat_basepaths(files, only_valid=False):
    """Returns unique basepaths from IDAT files or directory.

    This function processes a list of IDAT files or a directory containing IDAT
    files and returns their basepaths by removing the file endings. The
    function ensures that there are no duplicate basepaths in the returned list
    and maintains the order of the files as they appear in the input.

    Args:
        files (path or list): A file or directory path or a list of file paths.
        only_valid (bool): If True, only returns basepaths that point to valid
            IDAT file pairs. Defaults is 'False'.

    Returns:
        list: A list of unique basepaths corresponding to the provided IDAT
            files. If a directory is provided, all IDAT files are recursively
            considered.

    Example:
        >>> idat_basepaths("/path/to/dir")
        [PosixPath('/path/to/dir/file1'), PosixPath('/path/to/dir/file2')]

        >>> idat_basepaths(["/path1/file1_Grn.idat", "/path2/file2_Red.idat"])
        [PosixPath('/path1/file1'), PosixPath('/path2/file2')]

        >>> idat_basepaths("/path/to/idat/file_Grn.idat.gz")
        [PosixPath('/path/to/idat/file')]
    """

    def get_idat_files(file_or_dir):
        path = os.path.expanduser(file_or_dir)
        # If path is dir take all files in it
        if os.path.isdir(path):
            for dirpath, _, filenames in os.walk(path, followlinks=True):
                for filename in filenames:
                    if filename.endswith(ENDING_SUFFIXES):
                        yield os.path.join(dirpath, filename)
        else:
            yield path

    def strip_suffix(file_path):
        for suffix in ENDING_SUFFIXES:
            if file_path.endswith(suffix):
                return file_path[: -len(suffix)]
        return file_path

    if not isinstance(files, list):
        files = [files]
    _files = [
        strip_suffix(idat_file)
        for file_or_dir in files
        for idat_file in get_idat_files(file_or_dir)
    ]
    # Remove duplicates, keep ordering
    unique_basepaths_dict = dict.fromkeys(_files)
    if only_valid:
        return [
            Path(base)
            for base in unique_basepaths_dict
            if is_valid_idat_basepath(base)
        ]
    return [Path(base) for base in unique_basepaths_dict]


def idat_paths_from_basenames(basenames):
    """Returns paths to green and red IDAT files.

    Args:
        basenames (list): List of basepaths for IDAT files.

    Returns:
        tuple: Paths to green and red IDAT files.

    Raises:
        FileNotFoundError: If any IDAT file is not found.
    """
    grn_idat_files = np.array(
        [Path(str(name) + ENDING_GRN) for name in basenames]
    )
    red_idat_files = np.array(
        [Path(str(name) + ENDING_RED) for name in basenames]
    )

    def check_and_fix(files):
        not_existing = [i for i, path in enumerate(files) if not path.exists()]
        files[not_existing] = [
            x.parent / (x.name + ENDING_GZ) for x in files[not_existing]
        ]
        return next((x for x in files[not_existing] if not x.exists()), None)

    not_found = check_and_fix(grn_idat_files)
    not_found = (
        check_and_fix(red_idat_files) if not_found is None else not_found
    )
    if not_found is not None:
        idat_file = str(not_found).replace(ENDING_GZ, "")
        msg = f"IDAT file not found: {idat_file}."
        raise FileNotFoundError(msg)
    return grn_idat_files, red_idat_files


class RawData:
    """Represents raw intensity data extracted from IDAT files.

    This class initializes with a list of basepaths to IDAT files and parses
    them to extract raw intensity data from the green and red channels.

    Args:
        basenames (list): List of basepaths to IDAT files.
        manifest (Manifest, optional): The manifest associated with the array.
            If not provided, it will be determined from the probe count.

    Attributes:
        array_type (str): Type of Illumina array.
        probes (list): List of probe names corresponding to the IDAT files.
        ids (array): Array of probe IDs.
        _grn (array): Array of raw intensity values from the green channel.
        _red (array): Array of raw intensity values from the red channel.

    Example:
        >>> idat_basepath0 = directory_path / "200925700125_R07C01"
        >>> idat_basepath1 = directory_path / "200925700133_R02C01_Grn.idat"
        >>> raw_data = RawData(idat_basepath0)
        >>> raw_data = RawData([idat_basepath0, idat_basepath1])
    """

    def __init__(self, basenames, *, manifest=None):
        self.manifest = manifest
        _basenames = idat_basepaths(basenames)

        self.probes = [path.name.replace(ENDING_GZ, "") for path in _basenames]

        grn_idat_files, red_idat_files = idat_paths_from_basenames(_basenames)

        grn_idat = [
            IdatParser(str(filepath), intensity_only=True)
            for filepath in grn_idat_files
        ]
        red_idat = [
            IdatParser(str(filepath), intensity_only=True)
            for filepath in red_idat_files
        ]

        array_types = [
            ArrayType.from_probe_count(len(idat.illumina_ids))
            for idat in grn_idat + red_idat
        ]

        if len(set(array_types)) != 1:
            msg = "Array types must all be the same."
            raise ValueError(msg)

        self.array_type = array_types[0]

        all_illumina_ids = [idat.illumina_ids for idat in grn_idat + red_idat]

        if all(
            np.array_equal(all_illumina_ids[0], arr)
            for arr in all_illumina_ids
        ):
            self.ids = all_illumina_ids[0]
            self._grn = np.array([idat.probe_means for idat in grn_idat])
            self._red = np.array([idat.probe_means for idat in red_idat])
        else:
            self.ids = reduce(
                np.intersect1d, [idat.illumina_ids for idat in grn_idat]
            )
            self._grn = np.array(
                [
                    idat.probe_means[
                        np.isin(
                            idat.illumina_ids, self.ids, assume_unique=True
                        )
                    ]
                    for idat in grn_idat
                ]
            )
            self._red = np.array(
                [
                    idat.probe_means[
                        np.isin(
                            idat.illumina_ids, self.ids, assume_unique=True
                        )
                    ]
                    for idat in red_idat
                ]
            )
        if self.manifest is None:
            self.manifest = Manifest(self.array_type)

        self._grn_df = None
        self._red_df = None
        self.methylated = None
        self.unmethylated = None

    @property
    def grn(self):
        """DataFrame: Green channel raw intensity indexed by probe IDs."""
        if self._grn_df is None:
            self._grn_df = pd.DataFrame(
                self._grn.T, index=self.ids, columns=self.probes, dtype="int32"
            )
        return self._grn_df

    @property
    def red(self):
        """DataFrame: Red channel raw intensity indexed by probe IDs."""
        if self._red_df is None:
            self._red_df = pd.DataFrame(
                self._red.T, index=self.ids, columns=self.probes, dtype="int32"
            )
        return self._red_df

    def __repr__(self):
        title = "RawData():"
        lines = [
            title + "\n" + "*" * len(title),
            f"array_type: {self.array_type}",
            f"manifest: {self.manifest.array_type}",
            f"probes:\n{self.probes}",
            f"ids:\n{self.ids}",
            f"_grn:\n{self._grn}",
            f"_red:\n{self._red}",
            f"grn:\n{self.grn}",
            f"red:\n{self.red}",
        ]
        return "\n\n".join(lines)


@memoize
def _overlap_indices(left_arr, right_arr):
    """Compute the indices of overlapping elements between two arrays.

    This function finds the common elements (indices) between two input arrays
    and returns their positions in both arrays. It uses pandas Index objects
    and memoization for performance improvement.

    Example:
        >>> left_arr = ['a', 'b', 'c', 'd']
        >>> right_arr = ['b', 'c', 'e']
        >>> left_idx, right_idx = _overlap_indices(left_arr, right_arr)
        >>> print(left_idx)
        [1 2]
        >>> print(right_idx)
        [0 1]
    """
    if not isinstance(left_arr, pd.Index):
        left_arr = pd.Index(left_arr)
    if not isinstance(right_arr, pd.Index):
        right_arr = pd.Index(right_arr)
    common_indices = left_arr.intersection(right_arr)
    left_index = left_arr.get_indexer(common_indices)
    right_index = right_arr.get_indexer(common_indices)
    return left_index, right_index


class MethylData:
    """Represents methylated and unmethylated intensity data from RawData.

    This class provides methods for preprocessing Illumina methylation data and
    computing beta values from methylated and unmethylated intensities.

    Args:
        data (RawData): RawData object containing raw intensity data.
        file (str): Path to file or dir or list of paths containing raw
            intensity data.
        prep (str): Preprocessing method. Options: "illumina", "swan",
            "noob".
        seed (int, optional): Seed value used for random number generation in
            the SWAN preprocessing method. Default is None.

    Note:
        If 'data' is not provided, it will attempt to create a RawData object
        using the specified 'file'.

    Raises:
        ValueError: If neither 'data' nor 'file' is provided.
        ValueError: If 'data' is provided but is not of type 'RawData'.

    Examples:
        >>> methyl_data = MethylData(raw_data)
        >>> methyl_data = MethylData(file=file_path, prep="swan")
    """

    def __init__(self, data=None, file=None, prep="illumina", seed=None):
        if data is None and file is None:
            msg = "'data' or 'file' must be given."
            raise ValueError(msg)
        if data is None:
            data = RawData(file)
        elif not isinstance(data, RawData):
            msg = "'data' is not of type 'RawData'."
            raise ValueError(msg)
        self.seed = seed
        self._grn = data._grn
        self._red = data._red
        self.array_type = data.array_type
        self.ids = data.ids
        self.manifest = data.manifest
        self.probes = data.probes
        self.data = data
        self._grn_df = None
        self._red_df = None
        self._methylated_df = None
        self._unmethylated_df = None
        if prep == "illumina":
            self.preprocess_illumina()
        elif prep == "swan":
            self.preprocess_swan()
        elif prep == "noob":
            self.preprocess_noob()
        elif prep == "raw":
            self.preprocess_raw()
        else:
            msg = f"invalid 'prep' value {prep}"
            raise ValueError(msg)

    @property
    def grn(self):
        """DataFrame: Normalized green intensity by probe ID."""
        if self._grn_df is None:
            self._grn_df = pd.DataFrame(
                self._grn.T,
                index=self.ids,
                columns=self.probes,
                dtype="float32",
            )
        return self._grn_df

    @property
    def red(self):
        """DataFrame: Normalized red intensity by probe ID."""
        if self._red_df is None:
            self._red_df = pd.DataFrame(
                self._red.T,
                index=self.ids,
                columns=self.probes,
                dtype="float32",
            )
        return self._red_df

    @property
    def methylated(self):
        """DataFrame: Methylated intensity values indexed by IlmnID."""
        if self._methylated_df is None:
            self._methylated_df = pd.DataFrame(
                self.methyl.T,
                index=self.methyl_ilmnid,
                columns=self.probes,
                dtype="float32",
            )
            self._methylated_df.index.name = "IlmnID"
        return self._methylated_df

    @property
    def unmethylated(self):
        """DataFrame: Unmethylated intensity values indexed by IlmnID."""
        if self._unmethylated_df is None:
            self._unmethylated_df = pd.DataFrame(
                self.unmethyl.T,
                index=self.methyl_ilmnid,
                columns=self.probes,
                dtype="float32",
            )
            self._unmethylated_df.index.name = "IlmnID"
        return self._unmethylated_df

    def preprocess_illumina(self):
        """Performs preprocessing usings Illuminas method.

        This function implements preprocessing for Illumina methylation
        microarrays as used in Genome Studio, the standard software provided by
        Illumina.

        Details:
            This implementation is adapted from 'minfi'.
        """
        self._methylated_df = None
        self._unmethylated_df = None
        ci = MethylData._cached_indices(self.manifest, self.ids, "illumina")
        self._illumina_control_normalization(ci=ci)
        self._illumina_bg_correction(ci)
        self._preprocess_raw(ci)

    def _illumina_control_normalization(self, ci, reference=0):
        """Performs normalization using control probes."""
        grn_average = np.mean(
            self._grn[:, ci["ids_cg_cont"]],
            axis=1,
        )
        red_average = np.mean(
            self._red[:, ci["ids_at_cont"]],
            axis=1,
        )

        ref = (grn_average + red_average)[reference] / 2
        grn_factor = ref / grn_average
        red_factor = ref / red_average

        self._grn = grn_factor[:, np.newaxis] * self._grn
        self._red = red_factor[:, np.newaxis] * self._red

    def _illumina_bg_correction(self, ci):
        """Performs background normalization using negative control probes."""
        if len(ci["ids_ng_cont"]) <= 30:
            return

        grn_bg = np.partition(self._grn[:, ci["ids_ng_cont"]], 30)[:, 30]
        red_bg = np.partition(self._red[:, ci["ids_ng_cont"]], 30)[:, 30]

        # Subtract and threshold at zero, using in-place operations
        np.subtract(self._grn, grn_bg[:, np.newaxis], out=self._grn)
        np.maximum(self._grn, 0, out=self._grn)

        # Subtract and threshold at zero, using in-place operations
        np.subtract(self._red, red_bg[:, np.newaxis], out=self._red)
        np.maximum(self._red, 0, out=self._red)

    def _preprocess_raw_uncached(self):
        """Calculates methylated/unmethylated arrays without preprocessing.

        Converts the Red/Green channel for an Illumina methylation array
        into methylation signal, without using any normalization.

        Note:
            Uncached, slower version of ``preprocess_raw``.
        """
        type_1 = self.manifest.probe_info(ProbeType.ONE)
        type_2 = self.manifest.probe_info(ProbeType.TWO)
        type_1_red = type_1[type_1.Color_Channel.values == Channel.RED.value]
        type_1_grn = type_1[type_1.Color_Channel.values == Channel.GRN.value]
        man_idx_np = np.sort(
            np.concatenate(
                [
                    type_1.IlmnID.index,
                    type_2.IlmnID.index,
                ]
            )
        )
        self._preprocess_raw_methylated(
            man_idx_np, type_1_grn, type_1_red, type_2
        )
        self._preprocess_raw_unmethylated(
            man_idx_np, type_1_grn, type_1_red, type_2
        )

    def _preprocess_raw_methylated(self, man_idx_np, t1_grn, t1_red, t2):
        """Calculates methylated data frame without preprocessing."""
        red = self.red
        grn = self.grn
        result = pd.DataFrame(
            np.nan,
            index=man_idx_np,
            columns=red.columns,
            dtype="float32",
        )
        result.loc[t1_red.index] = red.loc[t1_red["AddressB_ID"].values].values
        result.loc[t1_grn.index] = grn.loc[t1_grn["AddressB_ID"].values].values
        result.loc[t2.index] = grn.loc[t2["AddressA_ID"].values].values
        result["IlmnID"] = self.manifest.data_frame.IlmnID.values[man_idx_np]
        self._methylated_df = result.set_index("IlmnID")

    def _preprocess_raw_unmethylated(self, man_idx_np, t1_grn, t1_red, t2):
        """Calculates unmethylated data frame without preprocessing."""
        red = self.red
        grn = self.grn
        result = pd.DataFrame(
            np.nan,
            index=man_idx_np,
            columns=red.columns,
            dtype="float32",
        )
        result.loc[t1_red.index] = red.loc[t1_red["AddressA_ID"].values].values
        result.loc[t1_grn.index] = grn.loc[t1_grn["AddressA_ID"].values].values
        result.loc[t2.index] = red.loc[t2["AddressA_ID"].values].values
        result["IlmnID"] = self.manifest.data_frame.IlmnID.values[man_idx_np]
        self._unmethylated_df = result.set_index("IlmnID")

    @memoize
    def _cached_indices(manifest, illumina_ids, prep):
        """Cache the indices required for data processing.

        Args:
            manifest (Manifest): Manifest object.
            illumina_ids (array): Array of Illumina IDs.
            prep (str): Preprocessing method. Options: "illumina", "noob",
                "swan", "raw".

        Returns:
            dict: Cached indices including probe indices, Illumina IDs indices,
                and probe type indices.
        """
        type_1 = manifest.probe_info(ProbeType.ONE)
        type_2 = manifest.probe_info(ProbeType.TWO)
        type_1_red = type_1[type_1.Color_Channel.values == Channel.RED.value]
        type_1_grn = type_1[type_1.Color_Channel.values == Channel.GRN.value]
        idx = pd.Index(
            np.sort(
                np.concatenate(
                    [
                        type_1.IlmnID.index,
                        type_2.IlmnID.index,
                    ]
                )
            )
        )
        ilmnid = manifest.data_frame.IlmnID.values[idx.values]
        ids = pd.Index(illumina_ids)
        ci = {"idx": idx.values, "ilmnid": ilmnid}
        ci["ids_1_red_a"] = ids.get_indexer(type_1_red["AddressA_ID"])
        ci["ids_1_red_b"] = ids.get_indexer(type_1_red["AddressB_ID"])
        ci["ids_1_grn_a"] = ids.get_indexer(type_1_grn["AddressA_ID"])
        ci["ids_1_grn_b"] = ids.get_indexer(type_1_grn["AddressB_ID"])
        ci["ids_2_____a"] = ids.get_indexer(type_2["AddressA_ID"])
        ci["idx_1_red__"] = idx.get_indexer(type_1_red.index)
        ci["idx_1_grn__"] = idx.get_indexer(type_1_grn.index)
        ci["idx_2______"] = idx.get_indexer(type_2.index)

        if prep == "illumina":
            at_controls = manifest.control_address(["NORM_A", "NORM_T"])
            ng_controls = manifest.control_address("NEGATIVE")
            cg_controls = manifest.control_address(["NORM_C", "NORM_G"])

            def valid_ids(indices):
                return indices[indices != -1]

            ci["ids_at_cont"] = valid_ids(ids.get_indexer(at_controls))
            ci["ids_cg_cont"] = valid_ids(ids.get_indexer(cg_controls))
            ci["ids_ng_cont"] = valid_ids(ids.get_indexer(ng_controls))

        if prep == "swan":
            ng_controls = manifest.control_address("NEGATIVE")

            def valid_ids(indices):
                return indices[indices != -1]

            ci["ids_ng_cont"] = valid_ids(ids.get_indexer(ng_controls))

        if prep == "noob":
            control_probes = manifest.control_data_frame
            control_probes = control_probes[
                control_probes.Address_ID.isin(ids)
            ].reset_index(drop=True)
            ci["idx_cg"] = control_probes[
                control_probes.Control_Type.isin(["NORM_C", "NORM_G"])
            ].index.values
            ci["idx_at"] = control_probes[
                control_probes.Control_Type.isin(["NORM_A", "NORM_T"])
            ].index.values
            ci["ids_ctr"] = ids.get_indexer(control_probes["Address_ID"])

        return ci

    def preprocess_raw(self):
        """Calculates methylated/unmethylated arrays without preprocessing.

        Converts the Red/Green channel for an Illumina methylation array
        into methylation signal, without using any normalization.
        """
        ci = MethylData._cached_indices(self.manifest, self.ids, "raw")
        self._preprocess_raw(ci)

    def _preprocess_raw_old(self, ci):
        """Same as _preprocess_raw just slower and cleaner."""
        self.methyl = np.full((len(self.probes), len(ci["idx"])), np.nan)
        self.methyl[:, ci["idx_1_red__"]] = self._red[:, ci["ids_1_red_b"]]
        self.methyl[:, ci["idx_1_grn__"]] = self._grn[:, ci["ids_1_grn_b"]]
        self.methyl[:, ci["idx_2______"]] = self._grn[:, ci["ids_2_____a"]]
        self.unmethyl = np.full((len(self.probes), len(ci["idx"])), np.nan)
        self.unmethyl[:, ci["idx_1_red__"]] = self._red[:, ci["ids_1_red_a"]]
        self.unmethyl[:, ci["idx_1_grn__"]] = self._grn[:, ci["ids_1_grn_a"]]
        self.unmethyl[:, ci["idx_2______"]] = self._red[:, ci["ids_2_____a"]]
        self.methyl_index = ci["idx"]
        self.methyl_ilmnid = ci["ilmnid"]

    def _preprocess_raw(self, ci):
        """Internal preprocess logic."""
        methyl_shape = (len(self.probes), len(ci["idx"]))
        self.methyl = np.full(methyl_shape, np.nan)
        self.unmethyl = np.full(methyl_shape, np.nan)
        self.methyl[:, ci["idx_1_red__"]] = np.take(
            self._red, ci["ids_1_red_b"], axis=1
        )
        self.methyl[:, ci["idx_1_grn__"]] = np.take(
            self._grn, ci["ids_1_grn_b"], axis=1
        )
        self.methyl[:, ci["idx_2______"]] = np.take(
            self._grn, ci["ids_2_____a"], axis=1
        )
        self.unmethyl[:, ci["idx_1_red__"]] = np.take(
            self._red, ci["ids_1_red_a"], axis=1
        )
        self.unmethyl[:, ci["idx_1_grn__"]] = np.take(
            self._grn, ci["ids_1_grn_a"], axis=1
        )
        self.unmethyl[:, ci["idx_2______"]] = np.take(
            self._red, ci["ids_2_____a"], axis=1
        )
        self.methyl_index = ci["idx"]
        self.methyl_ilmnid = ci["ilmnid"]

    def _swan_bg_intensity(self, ci):
        """Intensity background normalization used for SWAN preprocessing."""
        grn_med = np.median(
            self._grn[:, ci["ids_ng_cont"]],
            axis=1,
        )
        red_med = np.median(
            self._red[:, ci["ids_ng_cont"]],
            axis=1,
        )
        return np.mean([grn_med, red_med], axis=0)

    @staticmethod
    def _swan_indices(manifest, methyl_index, seed=None):
        rng = np.random.default_rng(seed)
        all_ncpgs = (
            manifest.data_frame[["Probe_Type", "N_CpG"]]
            .loc[methyl_index]
            .reset_index(drop=True)
        )
        subset_sizes = all_ncpgs.groupby(
            ["Probe_Type", "N_CpG"], dropna=False
        ).size()
        subset_size = min(
            subset_sizes.get((probe_type, n_cpg), 0)
            for probe_type in [ProbeType.ONE, ProbeType.TWO]
            for n_cpg in [1, 2, 3]
        )
        all_indices = {}
        random_subset_indices = {}
        for probe_type in [ProbeType.ONE, ProbeType.TWO]:
            all_ncpts_type = all_ncpgs[all_ncpgs.Probe_Type == probe_type]
            all_indices[probe_type] = all_ncpts_type.index.values
            all_ncpts_type = all_ncpts_type.reset_index(drop=True)
            indices = []
            for ncpgs in range(1, 4):
                ids = all_ncpts_type.index[all_ncpts_type.N_CpG == ncpgs]
                ids_subset = rng.permutation(ids)[:subset_size]
                indices.append(ids_subset)
            random_subset_indices[probe_type] = np.sort(
                np.concatenate(indices)
            )
        return all_indices, random_subset_indices

    def preprocess_swan(self):
        """Subset-quantile Within Array Normalization (SWAN).

        Details:
            The SWAN method has two parts. First, an average quantile
            distribution is created using a subset of probes defined to be
            biologically similar based on the number of CpGs underlying the
            probe body. This is achieved by randomly selecting N Infinium I and
            II probes that have 1, 2 and 3 underlying CpGs, where N is the
            minimum number of probes in the 6 sets of Infinium I and II probes
            with 1, 2 or 3 probe body CpGs. This results in a pool of 3N
            Infinium I and 3N Infinium II probes. The subset for each probe
            type is then sorted by increasing intensity.  The value of each of
            the 3N pairs of observations is subsequently assigned to be the
            mean intensity of the two probe types for that row or “quantile”.
            This is the standard quantile procedure. The intensities of the
            remaining probes are then separately adjusted for each probe type
            using linear interpolation between the subset probes.


            Implementation adapted from 'minfi'

        Note:
            SWAN uses a random subset of probes for between array
            normalization. To achieve reproducible results, set the seed.

        References:
            J Maksimovic, L Gordon and A Oshlack (2012). SWAN: Subset quantile
            Within-Array Normalization for Illumina Infinium
            HumanMethylation450 BeadChips. Genome Biology 13, R44.
        """
        self._methylated_df = None
        self._unmethylated_df = None
        ci = MethylData._cached_indices(self.manifest, self.ids, "swan")
        self._preprocess_raw(ci)
        bg_intensity = self._swan_bg_intensity(ci)
        all_indices, random_subset_indices = MethylData._swan_indices(
            self.manifest, self.methyl_index, self.seed
        )
        self.methyl = MethylData._preprocess_swan_main(
            self.methyl, bg_intensity, all_indices, random_subset_indices
        )
        self.unmethyl = MethylData._preprocess_swan_main(
            self.unmethyl, bg_intensity, all_indices, random_subset_indices
        )

    @staticmethod
    def _preprocess_swan_main(
        intensity, bg_intensity, all_indices, random_subset_indices
    ):
        """Main function for preprocess_swan."""
        from scipy.stats import rankdata

        random_subset_one = all_indices[ProbeType.ONE][
            random_subset_indices[ProbeType.ONE]
        ]
        random_subset_two = all_indices[ProbeType.TWO][
            random_subset_indices[ProbeType.TWO]
        ]
        sorted_subset_intensity = (
            np.sort(intensity[:, random_subset_one], axis=1)
            + np.sort(intensity[:, random_subset_two], axis=1)
        ) / 2
        swan = np.full(intensity.shape, np.nan)
        for i in range(len(intensity)):
            for probe_type in [ProbeType.ONE, ProbeType.TWO]:
                curr_intensity = intensity[i, all_indices[probe_type]]
                x = rankdata(curr_intensity) / len(curr_intensity)
                xp = np.sort(x[random_subset_indices[probe_type]])
                fp = sorted_subset_intensity[i, :]
                interp = np.interp(x=x, xp=xp, fp=fp)
                intensity_min = np.min(
                    curr_intensity[random_subset_indices[probe_type]]
                )
                intensity_max = np.max(
                    curr_intensity[random_subset_indices[probe_type]]
                )
                interp[x > np.max(xp)] += (
                    curr_intensity[x > np.max(xp)] - intensity_max
                )
                interp[x < np.min(xp)] += (
                    curr_intensity[x < np.min(xp)] - intensity_min
                )
                interp[interp <= 0] = bg_intensity[i]
                swan[i, all_indices[probe_type]] = interp
        return swan

    def preprocess_noob(self, offset=15, dye_method="single"):
        """The Noob preprocessing method.

        Description:
            Noob (normal-exponential out-of-band) is a background correction
            method with dye-bias normalization.

        Args:
            offset (float): An offset for the normexp background correction.
            dye_method (str): How should dye bias correction be done: "single"
                for single sample approach, or "reference" for a reference
                array.

        References:
            TJ Triche, DJ Weisenberger, D Van Den Berg, PW Laird and KD
            Siegmund _Low-level processing of Illumina Infinium DNA
            Methylation BeadArrays.  Nucleic Acids Res (2013) 41, e90.
            doi:10.1093/nar/gkt090.
        """
        self._methylated_df = None
        self._unmethylated_df = None

        ci = MethylData._cached_indices(self.manifest, self.ids, "noob")

        self._preprocess_raw(ci)

        grn_oob = np.concatenate(
            [self._grn[:, ci["ids_1_red_a"]], self._grn[:, ci["ids_1_red_b"]]],
            axis=1,
        )
        red_oob = np.concatenate(
            [self._red[:, ci["ids_1_grn_a"]], self._red[:, ci["ids_1_grn_b"]]],
            axis=1,
        )

        methyl = self.methyl
        unmethyl = self.unmethyl
        methyl[methyl <= 0] = 1
        unmethyl[unmethyl <= 0] = 1

        grn_m = methyl[:, ci["idx_1_grn__"]]
        grn_u = unmethyl[:, ci["idx_1_grn__"]]
        grn_2 = methyl[:, ci["idx_2______"]]

        xf_grn = np.concatenate([grn_m, grn_u, grn_2], axis=1)
        xs_grn = normexp_get_xs(xf_grn, controls=grn_oob, offset=offset)

        cumsum = np.cumsum([0, grn_m.shape[1], grn_u.shape[1], grn_2.shape[1]])
        slice_grn_m = slice(cumsum[0], cumsum[1])
        slice_grn_u = slice(cumsum[1], cumsum[2])
        slice_grn_2 = slice(cumsum[2], cumsum[3])

        red_m = methyl[:, ci["idx_1_red__"]]
        red_u = unmethyl[:, ci["idx_1_red__"]]
        red_2 = unmethyl[:, ci["idx_2______"]]

        xf_red = np.concatenate([red_m, red_u, red_2], axis=1)
        xs_red = normexp_get_xs(xf_red, controls=red_oob, offset=offset)

        cumsum = np.cumsum([0, red_m.shape[1], red_u.shape[1], red_2.shape[1]])
        slice_red_m = slice(cumsum[0], cumsum[1])
        slice_red_u = slice(cumsum[1], cumsum[2])
        slice_red_2 = slice(cumsum[2], cumsum[3])

        methyl[:, ci["idx_1_grn__"]] = xs_grn["xs"][:, slice_grn_m]
        unmethyl[:, ci["idx_1_grn__"]] = xs_grn["xs"][:, slice_grn_u]

        methyl[:, ci["idx_1_red__"]] = xs_red["xs"][:, slice_red_m]
        unmethyl[:, ci["idx_1_red__"]] = xs_red["xs"][:, slice_red_u]

        methyl[:, ci["idx_2______"]] = xs_grn["xs"][:, slice_grn_2]
        unmethyl[:, ci["idx_2______"]] = xs_red["xs"][:, slice_red_2]

        # Dye correction

        grn_control = self._grn[:, ci["ids_ctr"]]
        red_control = self._red[:, ci["ids_ctr"]]

        xcs_grn = normexp_get_xs(
            grn_control, param=xs_grn["param"], offset=offset
        )
        xcs_red = normexp_get_xs(
            red_control, param=xs_red["param"], offset=offset
        )

        grn_avg = np.mean(xcs_grn["xs"][:, ci["idx_cg"]], axis=1)
        red_avg = np.mean(xcs_red["xs"][:, ci["idx_at"]], axis=1)

        red_grn_ratio = red_avg / grn_avg

        if dye_method == "single":
            red_factor = 1 / red_grn_ratio
            grn_factor = np.array([1, 1])
        elif dye_method == "reference":
            ref_idx = np.argmin(np.abs(red_grn_ratio - 1))
            ref = (grn_avg + red_avg)[ref_idx] / 2
            if np.isnan(ref):
                msg = "'ref_idx' refers to an array that is not present"
                raise ValueError(msg)
            grn_factor = ref / grn_avg
            red_factor = ref / red_avg
        else:
            msg = "dye_method must be 'single' or 'reference'"
            raise ValueError(msg)

        red_factor = red_factor.reshape(-1, 1)
        methyl[:, ci["idx_1_red__"]] *= red_factor
        unmethyl[:, ci["idx_1_red__"]] *= red_factor
        unmethyl[:, ci["idx_2______"]] *= red_factor

        if dye_method == "reference":
            grn_factor = grn_factor.reshape(-1, 1)
            methyl[:, ci["idx_1_grn__"]] *= grn_factor
            unmethyl[:, ci["idx_1_grn__"]] *= grn_factor
            methyl[:, ci["idx_2______"]] *= grn_factor

        self.methyl = methyl
        self.unmethyl = unmethyl

    @property
    def betas(self):
        """Returns beta values."""
        betas = self._get_beta(self.methyl, self.unmethyl)
        return pd.DataFrame(
            betas.T, columns=self.probes, index=self.methyl_ilmnid
        )

    def betas_at(self, cpgs=None, fill=NEUTRAL_BETA):
        """Calculates beta values for specified CpG sites.

        Args:
            cpgs (array-like): Array of CpG IDs.
            fill (float): Value to fill for CpGs not found in the used
                manifest or equal to NaN.

        Returns:
            pandas.DataFrame: DataFrame containing beta values for specified
                CpGs.

        Nore:
            If 'cpgs' is None, all CpGs from the used manifest are considered.
        """
        if cpgs is None:
            cpgs = self.manifest.methylation_probes
        betas = self._get_beta(self.methyl, self.unmethyl)
        converted = np.full((len(self.probes), len(cpgs)), fill)
        left_idx, right_idx = _overlap_indices(cpgs, self.methyl_ilmnid)
        converted[:, left_idx] = betas[:, right_idx]
        converted[np.isnan(converted)] = fill
        return pd.DataFrame(converted.T, columns=self.probes, index=cpgs)

    @staticmethod
    def _get_beta(
        methylated, unmethylated, offset=0, beta_threshold=0, *, min_zero=True
    ):
        if offset < 0:
            msg = "'offset' must be non-negative"
            raise ValueError(msg)

        if not (0 <= beta_threshold <= 0.5):
            msg = "'beta_threshold' must be between 0 and 0.5"
            raise ValueError(msg)

        if min_zero:
            methylated = np.maximum(methylated, 0)
            unmethylated = np.maximum(unmethylated, 0)

        # Ignore division by zero
        with np.errstate(divide="ignore", invalid="ignore"):
            betas = methylated / (methylated + unmethylated + offset)

        if beta_threshold > 0:
            betas = np.minimum(
                np.maximum(betas, beta_threshold), 1 - beta_threshold
            )

        return betas

    def __repr__(self):
        title = "MethylData():"
        lines = [
            title + "\n" + "*" * len(title),
            f"array_type: {self.array_type}",
            f"manifest: {self.manifest.array_type}",
            f"probes:\n{self.probes}",
            f"_grn:\n{self._grn}",
            f"_red:\n{self._red}",
            f"grn:\n{self.grn}",
            f"red:\n{self.red}",
            f"methylated:\n{self.methylated}",
            f"unmethylated:\n{self.unmethylated}",
        ]
        if hasattr(self, "intensity"):
            lines.append(f"intensity:\n{self.intensity}")
        return "\n\n".join(lines)


class ReferenceMethylData:
    """Stores and manages reference cases for different array types.

    This class categorizes and processes reference IDAT files to create
    MethylData objects for different array types. It is intended for CNV
    neutral reference cases used in CNV calculation.

    Args:
        file (list): List of file paths to IDAT files or directory containing
            IDAT files.
        prep (str): Preprocessing method. Options: "illumina", "swan", "noob".

    Attributes:
        _methyl_data (dict): Internal dictionary to cache MethylData objects
            for each array type.

    Raises:
        ValueError: If no reference files are found for the specified array
            type.

    Examples:
        >>> # 'directory' contains 450k, EPIC and EPICv2 idat files
        >>> reference = ReferenceMethylData(file=directory, prep="illumina")
        >>> sample_450k = MethylData(file=idat_file_450k)
        >>> sample_epic = MethylData(file=idat_file_epic)
        >>> sample_epicv2 = MethylData(file=idat_file_epicv2)
        >>> # reference can be used for all types
        >>> cnv_450k = CNV(sample_450k, reference)
        >>> cnv_epic = CNV(sample_epic, reference)
        >>> cnv_epicv2 = CNV(sample_epicv2, reference)
    """

    _cache = {}

    def __new__(cls, file, prep="illumina", save_to_disk=False):
        key = cache_key(file, prep)
        if key in cls._cache:
            return cls._cache[key]

        instance = super().__new__(cls)

        # Cache the instance
        cls._cache[key] = instance
        return instance

    def __getnewargs__(self):
        # Necessary for pickle
        return self.file, self.prep, self.save_to_disk

    def __init__(self, file, prep="illumina", save_to_disk=False):
        # Don't need to initialize if instance is cached.
        if hasattr(self, "_cached"):
            return
        self._cached = True

        self.file = file
        self.prep = prep
        self.save_to_disk = save_to_disk
        idat_files = idat_basepaths(self.file)

        # Load data from disk
        filepath = ReferenceMethylData.pickle_filename(self.prep, idat_files)
        if self.save_to_disk and filepath.exists():
            with filepath.open("rb") as f:
                saved_instance = pickle.load(f)
                self.__dict__.update(saved_instance.__dict__)
                return

        reference_files = collections.defaultdict(list)
        self._methyl_data = {}
        for idat_file in tqdm(
            idat_files, desc="Categorizing reference IDAT files"
        ):
            array_type = ArrayType.from_idat(idat_file)
            reference_files[array_type].append(idat_file)
        for array_type, file_list in tqdm(
            reference_files.items(), desc="Processing reference IDAT files"
        ):
            raw_data = RawData(file_list)
            self._methyl_data[array_type] = MethylData(
                raw_data, prep=self.prep
            )

        # Save to disk
        if self.save_to_disk:
            with filepath.open("wb") as f:
                pickle.dump(self, f)

    @staticmethod
    def pickle_filename(prep, idat_files):
        return MEPYLOME_TMP_DIR / input_args_id(
            "Ref", prep, sorted(str(x) for x in idat_files)
        )

    def __getitem__(self, array_type):
        if array_type not in self._methyl_data:
            msg = (
                f"No copy number neutral reference files found for "
                f"array type '{array_type.value}'."
            )
            raise ValueError(msg)
        return self._methyl_data[array_type]



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.6
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
#
# <img alt="Mepylome Logo" src="https://raw.githubusercontent.com/brj0/mepylome/main/mepylome/data/assets/mepylome.svg" width="300">
#
# Mepylome: A Toolkit for DNA-Methylation Analysis in Tumor Diagnostics
# =====================================================================
#
# This notebook automates the analysis outlined in the Mepylome publication,
# performing all necessary steps from downloading datasets to executing the
# analyses.
#
#
# ### Usage
#
# - Follow the notebook/script step-by-step.
#
#
# ### System Tested
#
# - *Operating System*: Ubuntu 20.04.6
# - *Python Version*: 3.12
#
#
# ### Reference Publication (will follow)
#
# - *Authors*: Jon Brugger et al.
#
#
# ### Run This Notebook in Google Colab
#
# You can quickly open and run this notebook in Google Colab without any setup
# by clicking the link below.
#
# **Note**: The graphical user interface (GUI) features are limited in Google
# Colab. If using the free version, memory constraints may arise. Additionally,
# long download operations  may face timeouts or interruptions.
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/brj0/mepylome/blob/main/examples/publication/scc.ipynb)
#
#
# This notebook was automatically generated from the corresponding py-file
# with:
#
# ```bash
# jupytext --to ipynb *.py
# ```

# %% [markdown]
# -----------------------------------------------------------------------------
# ## Contents
# 0. **[Initialization](#0.-Initialization)**
# 1. **[Data Loading](#1.-Data-Loading)**
# 2. **[UMAP Calculation](#2.-UMAP-Calculation)**
# 3. **[Supervised Classifier Training](#3.-Supervised-Classifier-Training)**
# 4. **[CNV Analysis](#4.-CNV-Analysis)**


# %% [markdown]
# -----------------------------------------------------------------------------
# <a name="0.-Initialization"></a>
# ## 0. Initialization
#
# ### Install Required Packages
#
# To run the analysis, install the following Python packages:
# - `mepylome` for DNA-methylation analysis
# - `ruptures` for segmentation in CNV plots
# - `ipython`, `pillow`, and `ipywidgets` for interactive and graphical
#   functionality
# - `kaleido` for saving plots
#
# Install them (1-2 minutes) using:

# %% language="bash"
#
# pip install mepylome ruptures ipython pillow ipywidgets kaleido


# %% [markdown]
# ### Core Imports, Configuration and main Functions

# %%
import io
import json
import multiprocessing
import os
import platform
import re
import sys
import tarfile
import zipfile
from concurrent.futures import (
    ThreadPoolExecutor,
    as_completed,
)
from pathlib import Path
from threading import Lock

import numpy as np
import pandas as pd
import requests
from IPython.display import Image as IPImage
from PIL import Image

from mepylome import ArrayType, Manifest, clear_cache, idat_basepaths
from mepylome.analysis import MethylAnalysis
from mepylome.dtypes.manifests import (
    DOWNLOAD_DIR,
    MANIFEST_URL,
    REMOTE_FILENAME,
)
from mepylome.utils.files import (
    download_file,
    download_geo_samples,
)

# Define output font size for plots
FONTSIZE = 23
IMG_HEIGHT = 2000
IMG_WIDTH = 1000
GEO_URL = "https://www.ncbi.nlm.nih.gov/geo/download/?acc={acc}&format=file"

# Define dataset URLs and filenames
datasets = {
    "scc": {
        "xlsx": "https://www.science.org/doi/suppl/10.1126/scitranslmed.aaw8513/suppl_file/aaw8513_data_file_s1.xlsx",
        "geo_ids": [],
    },
    "scc_test": {
        "geo_ids": [
            "GSE124052",  # HNSQ_CA, NSCLC_SC
            "GSE66836",  # NSCLC_AD, CONTR_LUNG
            "GSE79556",  # HNSQ_CA (oral tongue)
            "GSE87053",  # HNSQ_CA, CONTR_OC
            "GSE95036",  # HNSQ_CA
            "GSE124052",  # HNSQ_CA, NSCLC_SC
        ],
    },
}

# Determine basic storage directory depending on platform
if "COLAB_GPU" in os.environ:
    # Google Colab
    mepylome_dir = Path("/content/mepylome")
elif Path("/mnt/bender").exists():
    # Bender-specific path
    mepylome_dir = Path("/mnt/bender/mepylome")
else:
    # Default for local Linux or other environments
    mepylome_dir = Path.home() / "mepylome"


data_dir = mepylome_dir / "data"
output_dir = mepylome_dir / "outputs"
reference_dir = mepylome_dir / "cnv_references"
validation_dir = mepylome_dir / "validation_data"

# Ensure the directory exists
mepylome_dir.mkdir(parents=True, exist_ok=True)
data_dir.mkdir(parents=True, exist_ok=True)
output_dir.mkdir(parents=True, exist_ok=True)
reference_dir.mkdir(parents=True, exist_ok=True)
validation_dir.mkdir(parents=True, exist_ok=True)


print("=== System Information ===")
print(f"Python Version: {sys.version.split()[0]}")
print(f"Platform: {platform.system()} {platform.release()}")
print(f"Processor: {platform.processor()}")
print(f"Number of CPUs: {multiprocessing.cpu_count()}")
print(f"Data will be stored in: {mepylome_dir}")


# Main Functions


def extract_tar(tar_path, output_directory):
    """Extracts tar file under 'tar_path' to 'output_directory'."""
    output_directory.mkdir(parents=True, exist_ok=True)
    with tarfile.open(tar_path, "r") as tar:
        tar.extractall(path=output_directory, filter="data")
        print(f"Extracted {tar_path} to {output_directory}")


def download_from_geo_and_untar(analysis_dir, geo_ids):
    """Downloads all missing GEO files and untars them."""
    for geo_id in geo_ids:
        idat_dir = analysis_dir / geo_id
        if idat_dir.exists():
            print(f"Data for GEO ID {geo_id} already exists, skipping.")
            continue
        try:
            tar_path = analysis_dir / f"{geo_id}.tar"
            geo_url = GEO_URL.format(acc=geo_id)
            download_file(geo_url, tar_path)
            extract_tar(tar_path, idat_dir)
            tar_path.unlink()
        except Exception as e:
            print(f"Error processing GEO ID {geo_id}: {e}")


def clean_filename(name):
    """Replace invalid characters with a single underscore."""
    return re.sub(r"[^\w\-]+", "_", name)


def calculate_cn_summary(analysis, class_):
    """Calculates and saves CN summary plots."""
    df_class = analysis.idat_handler.samples_annotated[class_]
    plot_list = []
    analysis_dir = analysis.analysis_dir
    all_classes = sorted(df_class.unique())
    for methyl_class in all_classes:
        df_index = df_class == methyl_class
        sample_ids = df_class.index[df_index]
        plot, _ = analysis.cn_summary(sample_ids)
        plot.update_layout(
            title=f"{methyl_class}",
            title_x=0.5,
            yaxis_title="Proportion of CNV gains/losses",
        )
        plot.update_layout(
            title_font_size=FONTSIZE + 3,
            yaxis_title_font_size=FONTSIZE - 2,
        )
        plot_list.append(plot)
    png_paths = [
        output_dir / f"{analysis_dir.name}_cn_summary_{clean_filename(x)}.png"
        for x in all_classes
    ]
    for path, fig in zip(png_paths, plot_list):
        fig.write_image(path)
    images = [Image.open(path) for path in png_paths]
    width, height = images[0].size
    n_columns = 4
    n_images = len(images)
    n_rows = (n_images + n_columns - 1) // n_columns
    total_width = width * n_columns
    total_height = height * n_rows
    new_image = Image.new("RGB", (total_width, total_height), (255, 255, 255))
    for index, img in enumerate(images):
        row = index // n_columns
        col = index % n_columns
        x = col * width
        y = row * height
        new_image.paste(img, (x, y))
    output_path = output_dir / f"{analysis_dir.name}_cn_summary.png"
    new_image.save(output_path)
    return output_path


# %% [markdown]
# ### Blacklist Generation for CpG Sites
#
# Some CpG sites should be excluded from the analysis. Here we choose probes
# flagged with `MFG_Change_Flagged` that should be excluded according to the
# manifest and those that are on sex chromosomes.


# %%
def generate_blacklist_cpgs():
    """Returns and caches CpG sites that should be blacklisted."""
    print("Generating blacklist. Can take some time...")
    blacklist_path = data_dir / "cpg_blacklist.csv"
    if not blacklist_path.exists():
        manifest_url = MANIFEST_URL[ArrayType.ILLUMINA_EPIC]
        DOWNLOAD_DIR.mkdir(parents=True, exist_ok=True)
        response = requests.get(manifest_url)
        html_sucess_ok_code = 200
        if response.status_code == html_sucess_ok_code:
            with zipfile.ZipFile(io.BytesIO(response.content)) as thezip:
                thezip.extractall(DOWNLOAD_DIR)
        else:
            msg = f"Failed to download the file: {response.status_code}"
            raise RuntimeError(msg)
        csv_path = DOWNLOAD_DIR / REMOTE_FILENAME[ArrayType.ILLUMINA_EPIC]
        manifest_df = pd.read_csv(csv_path, skiprows=7, low_memory=False)
        flagged_cpgs = manifest_df[
            manifest_df["MFG_Change_Flagged"].fillna(False)
        ]["IlmnID"]
        flagged_cpgs.to_csv(blacklist_path, index=False, header=False)
        csv_path.unlink()
    blacklist_df = pd.read_csv(blacklist_path, header=None)
    print("Generating blacklist done.")
    return set(blacklist_df.iloc[:, 0])


def sex_chromosome_cpgs():
    """Returns CpGs on sex chromosomes for EPIC and 450k arrays."""
    manifest = Manifest("epic")
    sex_cpgs_epic = manifest.data_frame[
        manifest.data_frame.Chromosome.isin([23, 24])
    ].IlmnID
    manifest = Manifest("450k")
    sex_cpgs_450k = manifest.data_frame[
        manifest.data_frame.Chromosome.isin([23, 24])
    ].IlmnID
    return set(sex_cpgs_epic) | set(sex_cpgs_450k)


# Choose CpG list that should be blacklisted
blacklist = generate_blacklist_cpgs() | sex_chromosome_cpgs()

# %% [markdown]
# ### CNV-Neutral Reference Samples
#
# For generating copy number variation (CNV) plots, a sufficiently large set of
# CNV-neutral reference samples is required. Here, we use control samples from
# [Koelsche et al. (2021)](https://doi.org/10.1038/s41467-020-20603-4). These
# samples are stored in the designated reference_dir.
#
# **Best Practices**:
# - Include both fresh-frozen and FFPE (formalin-fixed paraffin-embedded)
#   samples in the copy-neutral reference set for optimal results.

# %%
cn_neutral_samples = [
    "GSM4180453_201904410008_R06C01",
    "GSM4180454_201904410008_R05C01",
    "GSM4180455_201904410008_R04C01",
    "GSM4180456_201904410008_R03C01",
    "GSM4180457_201904410008_R02C01",
    "GSM4180458_201904410008_R01C01",
    "GSM4180459_201904410007_R08C01",
    "GSM4180460_201904410007_R07C01",
    "GSM4180741_201247480004_R05C01",
    "GSM4180742_201247480004_R04C01",
    "GSM4180743_201247480004_R03C01",
    "GSM4180751_201194010006_R01C01",
    "GSM4180909_200394870074_R04C02",
    "GSM4180910_200394870074_R03C02",
    "GSM4180911_200394870074_R02C02",
    "GSM4180912_200394870074_R01C02",
    "GSM4180913_200394870074_R05C01",
    "GSM4180914_200394870074_R04C01",
    "GSM4181456_203049640041_R03C01",
    "GSM4181509_203049640040_R07C01",
    "GSM4181510_203049640040_R08C01",
    "GSM4181511_203049640041_R01C01",
    "GSM4181512_203049640041_R02C01",
    "GSM4181513_203049640041_R04C01",
    "GSM4181514_203049640041_R05C01",
    "GSM4181515_203049640041_R06C01",
    "GSM4181516_203049640041_R07C01",
    "GSM4181517_203049640041_R08C01",
]

download_geo_samples(reference_dir, cn_neutral_samples)


# %% [markdown]
# -----------------------------------------------------------------------------
# <a name="1.-Data-Loading"></a>
# ## 1. Data Loading
#
# In this example, we aim to reproduce the pan-SCC classifier presented in the
# study by [Jurmeister et al.
# (2019)](https://doi.org/10.1126/scitranslmed.aaw8513). Our goal is to gather
# data for Squamous Cell Carcinoma (SCC) from multiple sources, as outlined in
# the publication, including datasets from The Cancer Genome Atlas (TCGA) and
# from the Gene Expression Omnibus (GEO) repository. Datasets without IDAT
# files are omitted from the collection process.

# %%
# Initialize directories.
tumor_site = "scc"
analysis_dir = data_dir / tumor_site
test_dir = validation_dir / tumor_site

test_dir.mkdir(parents=True, exist_ok=True)
analysis_dir.mkdir(parents=True, exist_ok=True)

# %% [markdown]
# ### Step 1: Download TCGA Data
#
# We download all SCC samples from TCGA that contain IDAT files. **This may
# take several hours.**
#
# **Important:** Downloading from TCGA can be unreliable. Connection resets and
# server-side interruptions are common.
#
# This script is designed to be resilient: it automatically **retries failed
# downloads** up to 10 times. After the final attempt, it will continue
# regardless of whether some files failed. You can re-run this cell to try
# downloading any remaining files.


# %%
tcga_dir = analysis_dir / "tcga_scc"
tcga_downloaded_tag = tcga_dir / ".download_complete"
tcga_metadata_dir_tar = analysis_dir / "tcga_metadata.tar.gz"
tcga_metadata_dir = tcga_metadata_dir_tar.with_suffix("").with_suffix("")

tcga_dir.mkdir(parents=True, exist_ok=True)

geo_metadata_url = "https://raw.githubusercontent.com/brj0/mepylome-data/main/examples/geo_metadata.tar.gz"
tcga_metadata_url = "https://raw.githubusercontent.com/brj0/mepylome-data/main/examples/tcga_metadata.tar.gz"

# Check if the TCGA annotation tar file exists and extract
if not tcga_metadata_dir.exists():
    print("Setting up TCGA annotation directory...")
    download_file(tcga_metadata_url, tcga_metadata_dir_tar)
    extract_tar(tcga_metadata_dir_tar, analysis_dir)
    print("Setting up TCGA annotation directory done.")


def download_tcga_file(file_id, filename, dest_folder):
    """Downloads a TCGA file safely, cleaning up incomplete downloads."""
    final_path = dest_folder / filename
    temp_path = final_path.with_suffix(final_path.suffix + ".part")
    if final_path.exists():
        print(f"File already exists: {filename}")
        return True
    url = f"https://api.gdc.cancer.gov/data/{file_id}"
    dest_folder.mkdir(parents=True, exist_ok=True)
    try:
        with requests.get(url, stream=True, timeout=60) as r:
            r.raise_for_status()
            with temp_path.open("wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
        temp_path.rename(final_path)
        return True
    except Exception as e:
        print(f"Failed: {filename} | Error: {e}")
        if temp_path.exists():
            temp_path.unlink()
        return False


def download_all_tcga_files(manifest_path, dest_folder, max_threads):
    """Downloads all TCGA files in manifest in parallel."""
    manifest = pd.read_csv(manifest_path, sep="\t")
    total_files = len(manifest)
    pending = [
        item
        for _, item in manifest.iterrows()
        if not (dest_folder / item["filename"]).exists()
    ]
    attempt = 1
    downloaded = total_files - len(pending)
    MAX_ATTEMPTS = 10
    lock = Lock()
    while pending and attempt <= MAX_ATTEMPTS:
        print(f"\n--- Attempt {attempt}: {len(pending)} files remaining ---")
        next_round = []
        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            futures = {
                executor.submit(
                    download_tcga_file,
                    item["id"],
                    item["filename"],
                    dest_folder,
                ): item
                for item in pending
            }
            for future in as_completed(futures):
                item = futures[future]
                success = future.result()
                if success:
                    with lock:
                        downloaded += 1
                    print(
                        f"Downloaded file {downloaded}/{total_files}: "
                        f"{item['filename']}"
                    )
                else:
                    next_round.append(item)
        pending = next_round
        attempt += 1
    if not pending:
        print("\n--- All files downloaded successfully ---")
        tcga_downloaded_tag.touch()
    else:
        print("\n--- Some files failed to download. Retry limit reached. ---")


# Check if the download is complete
if not tcga_downloaded_tag.exists():
    print("Download has not been completed yet.")
    print("Downloading TCGA files. This may take several hours!")
    manifest_file = next(tcga_metadata_dir.glob("gdc_manifest.*txt"))
    if not manifest_file.exists():
        msg = "No TCGA manifest file found."
        raise FileNotFoundError(msg)
    print(f"Downloading TCGA data from manifest file: {manifest_file}")
    download_all_tcga_files(manifest_file, tcga_dir, max_threads=8)
else:
    print("TCGA data already completely downloaded.")


# %% [markdown]
# Clean up by removing array types other than `450k` and `epic`.


# %%
def remove_invalid_array_types(root_dir):
    """Removes all IDAT files that are not of type 450k or epicv1."""
    idat_files = root_dir.glob("*idat")
    valid_array_types = {ArrayType.ILLUMINA_450K, ArrayType.ILLUMINA_EPIC}
    for idat_file in idat_files:
        array_type = ArrayType.from_idat(idat_file)
        if array_type not in valid_array_types:
            print(f"Removing {idat_file.name} (Type: {array_type})")
            idat_file.unlink()


remove_invalid_array_types(tcga_dir)


# %% [markdown]
# Next we extract the TCGA annotation.


# %%
def extract_tcga_case_id_dict(json_path):
    """Extracts a dictionary mapping from IDAT IDs to case IDs."""
    with json_path.open() as f:
        data = json.load(f)
    case_id_mapping = {}
    n_suffix = len("_Grn.idat")
    for item in data:
        file_name = item.get("file_name", "")[:-n_suffix]
        case_id = item.get("associated_entities", [{}])[0].get("case_id", "")
        if case_id and file_name:
            case_id_mapping[case_id] = file_name
    return case_id_mapping


json_metadata = next(tcga_metadata_dir.glob("metadata.cart.*json"))
case_id_to_sample_id = extract_tcga_case_id_dict(json_metadata)

# Load the clinical data and map the case_id to the IDAT
tcga_annotation = pd.read_csv(
    tcga_metadata_dir / "clinical.tsv", delimiter="\t"
)
tcga_annotation["Sample_ID"] = tcga_annotation["case_id"].map(
    case_id_to_sample_id
)
tcga_annotation = tcga_annotation.drop(columns="case_id")

# Rename columns restrict to the useful ones
columns_dict = {
    "gender": "Sex",
    "age_at_index": "Age",
    "tissue_or_organ_of_origin": "Tumor_site",
    "site_of_resection_or_biopsy": "Site_of_resection_or_biopsy",
    "tumor_grade": "Tumor_grade",
    "morphology": "Morphology",
    "primary_diagnosis": "Primary_diagnosis",
    "Sample_ID": "Sample_ID",
}
tcga_annotation = tcga_annotation.rename(columns=columns_dict)
tcga_annotation = tcga_annotation[columns_dict.values()].copy()

# Standardize the 'Sex' column and convert 'Age' to numeric
tcga_annotation["Sex"] = tcga_annotation["Sex"].replace(
    {"female": "Female", "male": "Male"}
)
tcga_annotation["Age"] = pd.to_numeric(tcga_annotation["Age"], errors="coerce")

# Mark the samples that to be censored
diag_to_censor_stat = {
    "Adenocarcinoma with mixed subtypes": 0,
    "Adenocarcinoma, NOS": 0,
    "Adenosquamous carcinoma": 1,
    "Basaloid squamous cell carcinoma": 1,
    "Lymphoepithelial carcinoma": 1,
    "Papillary carcinoma, NOS": 1,
    "Papillary squamous cell carcinoma": 1,
    "Squamous cell carcinoma, NOS": 0,
    "Squamous cell carcinoma, keratinizing, NOS": 0,
    "Squamous cell carcinoma, large cell, nonkeratinizing, NOS": 1,
    "Squamous cell carcinoma, nonkeratinizing, NOS": 0,
    "Squamous cell carcinoma, small cell, nonkeratinizing": 1,
    "Squamous cell carcinoma, spindle cell": 1,
    "Warty carcinoma": 1,
}

tcga_annotation["Censor"] = tcga_annotation["Primary_diagnosis"].map(
    diag_to_censor_stat
)

# Condense the primary tumor site.
nsclc_sites = {
    "Lower lobe, lung",
    "Lung, NOS",
    "Main bronchus",
    "Middle lobe, lung",
    "Overlapping lesion of lung",
    "Upper lobe, lung",
}
hnsq_sites = {
    "Anterior floor of mouth",
    "Base of tongue, NOS",
    "Border of tongue",
    "Cheek mucosa",
    "Floor of mouth, NOS",
    "Gum, NOS",
    "Hard palate",
    "Head, face or neck, NOS",
    "Hypopharynx, NOS",
    "Larynx, NOS",
    "Lip, NOS",
    "Lower gum",
    "Mandible",
    "Mouth, NOS",
    "Nasal cavity",
    "Oropharynx, NOS",
    "Overlapping lesion of lip, oral cavity and pharynx",
    "Palate, NOS",
    "Pharynx, NOS",
    "Posterior wall of oropharynx",
    "Retromolar area",
    "Supraglottis",
    "Tongue, NOS",
    "Tonsil, NOS",
    "Upper Gum",
    "Ventral surface of tongue, NOS",
}
cervix_sites = {"Cervix uteri"}
eso_sites = {
    "Cardia, NOS",
    "Esophagus, NOS",
    "Lower third of esophagus",
    "Middle third of esophagus",
    "Thoracic esophagus",
    "Upper third of esophagus",
}
censor_sites = {
    "Breast, NOS",
    "Bladder, NOS",
}

tcga_annotation["Diagnosis"] = None

# Classify each row based on the tumor site
for index, row in tcga_annotation.iterrows():
    site = str(row["Tumor_site"]).strip()
    diagnosis = row["Primary_diagnosis"]
    if diagnosis.startswith("Adenocarcinoma") and site in nsclc_sites:
        tcga_annotation.loc[index, "Diagnosis"] = "NSCLC_AD"
    elif site in cervix_sites:
        tcga_annotation.loc[index, "Diagnosis"] = "CERSQ_CA"
    elif site in nsclc_sites:
        tcga_annotation.loc[index, "Diagnosis"] = "NSCLC_SC"
    elif site in hnsq_sites:
        tcga_annotation.loc[index, "Diagnosis"] = "HNSQ_CA"
    elif site in eso_sites:
        tcga_annotation.loc[index, "Diagnosis"] = "ESO_CA_SQ"
    else:
        tcga_annotation.loc[index, "Censor"] = 1
        print(f"Unmatched tumor site: {site} (index {index} - censored)")

# Removed censored samples
tcga_annotation = tcga_annotation[tcga_annotation["Censor"] == 0]

# %% [markdown]
# ### Step 2: Download and unzip the GEO data

# %%
geo_metadata_dir_tar = analysis_dir / "geo_metadata.tar.gz"
geo_metadata_dir = geo_metadata_dir_tar.with_suffix("").with_suffix("")

# Check if the GEO annotation tar file exists and extract
if not geo_metadata_dir.exists():
    print("Setting up GEO annotation directory...")
    download_file(geo_metadata_url, geo_metadata_dir_tar)
    extract_tar(geo_metadata_dir_tar, analysis_dir)
    print("Setting up GEO annotation directory done.")

# Download the IDAT files.
download_from_geo_and_untar(analysis_dir, datasets[tumor_site]["geo_ids"])
download_from_geo_and_untar(
    test_dir, datasets[tumor_site + "_test"]["geo_ids"]
)


# Download the annotation spreadsheet.
def merge_csv(dir_path):
    """Reads all CSV files merges them."""
    dir_path = Path(dir_path)
    merged_df = pd.DataFrame()
    for csv_file in dir_path.glob("*.csv"):
        print(f"Reading {csv_file}")
        data_frame = pd.read_csv(csv_file)
        merged_df = pd.concat([merged_df, data_frame], ignore_index=True)
    return merged_df


geo_annotation = merge_csv(geo_metadata_dir)

# %% [markdown]
# ### Step 3: Construct the annotation file of all data.
# Join the TCGA and GEO annotation files

# %%
if (csv_path := analysis_dir / f"{tumor_site}.csv").exists():
    anno_df = pd.read_csv(csv_path)
    print("Merged annotation file allready exists.")
else:
    anno_df = pd.concat([geo_annotation, tcga_annotation], ignore_index=True)
    # Remove Adenocarcinoma and normal oral cavity samples
    scc_types = {"HNSQ_CA", "NSCLC_SC", "ESO_CA_SQ", "CERSQ_CA"}
    anno_df = anno_df[anno_df["Diagnosis"].isin(scc_types)]
    anno_df.to_csv(csv_path, index=False)
    print("Merged annotation file created.")


# %% [markdown]
# ### Create the Methylation Analysis Object
#
# The `MethylAnalysis` object serves as the main interface for performing DNA
# methylation analysis. Key parameters such as the directory structure, number
# of CpG sites, and UMAP settings are configured here.

# %%
# Only consider test files with annotation.
test_ids = set(anno_df.Sample_ID).intersection(
    x.name for x in idat_basepaths(test_dir)
)

analysis = MethylAnalysis(
    analysis_dir=analysis_dir,
    reference_dir=reference_dir,
    output_dir=output_dir,
    test_dir=test_dir,
    test_ids=test_ids,
    n_cpgs=25000,
    load_full_betas=True,
    overlap=True,
    cpg_blacklist=blacklist,
    cpgs="450k+epic+epicv2",
    debug=False,
    do_seg=True,
    umap_parms={
        "n_neighbors": 5,
        "metric": "manhattan",
        "min_dist": 0.1,
    },
)

# %% [markdown]
# ### Load Beta Values
#
# Reads and processes beta values from the provided dataset. This step is
# optional and primarily demonstrates the time required for processing. If not
# performed here, it will be automatically executed in the background when
# needed.

# %%
analysis.set_betas()


# %% [markdown]
# -----------------------------------------------------------------------------
# <a name="2.-UMAP-Calculation"></a>
# ## 2. UMAP Calculation
#
# ### Generate UMAP Plot
#
# Set the columns used for coloring the UMAP plot before initiating the
# dimensionality reduction process. The UMAP algorithm produces a visual
# representation of the sample clusters, which is stored as a Plotly object in
# `analysis.umap_plot`.

# %%
# Calculate UMAP
analysis.idat_handler.selected_columns = ["Diagnosis"]
analysis.make_umap()

# %%
# Show the results
print(analysis.umap_df)

# %%
# Generate and show image
output_path = output_dir / f"{analysis_dir.name}_umap_plot.jpg"
analysis.umap_plot.write_image(
    output_path,
    format="jpg",
    width=IMG_HEIGHT,
    height=IMG_WIDTH,
    scale=1,
)
IPImage(filename=output_path)


# %% [markdown]
# ### Launch the Analysis GUI
#
# Initializes an interactive GUI for further exploration of the methylation
# data.
#
# **Note:** This step works best in local environments and may have limitations
# on platforms like Google Colab or Binder.

# %%
analysis.run_app(open_tab=True)


# %% [markdown]
# On memory-limited platforms such as Google Colab, we need to manually free up
# memory between operations to avoid crashes.

# %%
# Free memory
clear_cache()


# %% [markdown]
# -----------------------------------------------------------------------------
# <a name="3.-Supervised-Classifier-Training"></a>
# ## 3. Supervised Classifier Training
#
# ### Supervised Classifier Validation
#
# The next step involves validating various supervised classification
# algorithms to evaluate their performance on the dataset. This process helps
# identify the most accurate model for methylation-based classification.
#
# **Note**:
# Training is resource- and time-intensive. The process may take up to 10
# minutes, depending on the computational resources available.

# %% [markdown]
# ***Preselect features (optional, recommended for low-memory systems)***:
#
# To optimize memory usage, we use a custom `feature_matrix` for training
# instead of the full `betas_all`, which includes all CpG sites. When the beta
# values are selected via `set_betas`, both `betas_sel` and `betas_all` are
# produced. While `betas_all` contains all CpGs, `betas_sel` (in our case)
# includes the top 25,000 most variable CpG sites.
#
# This step is crucial for low-memory systems (e.g., Google Colab) to reduce
# memory consumption. If memory is not a concern, you can skip this step and
# include all CpGs in the analysis.

# %%
# Preselect top 25,000 UMAP CpGs to reduce memory usage (optional)
analysis.feature_matrix = analysis.betas_sel


# %%
# Train supervised classifiers

analysis.idat_handler.selected_columns = ["Diagnosis"]
ids = analysis.idat_handler.ids
clf_out = analysis.classify(
    ids=ids,
    clf_list=[
        "vtl-kbest(k=10000)-et",
        "vtl-kbest(k=10000)-lr(max_iter=10000)",
        "vtl-kbest(k=10000)-rf",
    ],
)

# %%
# Print reports for all classifier for the first sample
for clf_result in clf_out:
    print(clf_result.reports["txt"][0])
    print()

# %%
# Identify and display the best classifier
best_clf = max(
    clf_out, key=lambda result: np.mean(result.metrics["accuracy_scores"])
)
print("Most accurate classifier:")
print(best_clf.reports["txt"][0])

# %% [markdown]
# Now we apply the best classifier on the independent validation samples:

# %%
test_ids = analysis.idat_handler.test_ids
# Ignore all files that are not in annotation file
test_ids = list(
    set(test_ids).intersection(analysis.idat_handler.annotation_df.index)
)
clf_out_pred = analysis.classify(ids=test_ids, clf_list=best_clf.model)
pred = clf_out_pred[0].prediction.idxmax(axis=1)
true_values = analysis.idat_handler.samples_annotated.loc[test_ids][
    "Diagnosis"
]
correct = np.sum(pred == true_values)
total = len(pred)
accuracy = correct / total
print(f"Classifier Accuracy: {100 * accuracy:.2f} % ({correct}/{total})")

# Analyze misclassified samples
misclassified = pred != true_values
misclassified_samples = clf_out_pred[0].prediction[misclassified].copy()
misclassified_samples["Pred"] = pred[misclassified]
misclassified_samples["True"] = true_values[misclassified]
print("Missclassified samples:\n", misclassified_samples)

# %%
# Free memory
clear_cache()


# %% [markdown]
# -----------------------------------------------------------------------------
# <a name="4.-CNV-Analysis"></a>
# ## 4. CNV Analysis
#
# %% [markdown]
# ### Generate and Save CNV Plot
#
# Creates a copy number variation (CNV) plot for a specified sample and saves
# the output as an image.

# %%
# Save CNV example
analysis.make_cnv_plot("GSM3519735_202915460126_R05C01")
cnv_plot = analysis.cnv_plot
cnv_plot.update_layout(
    yaxis_range=[-1.1, 1.1],
    font={"size": FONTSIZE},
    margin={"t": 50},
)
output_path = output_dir / f"{analysis_dir.name}_cnv_plot.jpg"
cnv_plot.write_image(
    output_path,
    format="jpg",
    width=IMG_HEIGHT,
    height=IMG_WIDTH,
    scale=1,
)
IPImage(filename=output_path)

# %% [markdown]
# ### Generate CNV Summary Plots
#
# In addition to individual CNV plots, this step computes summary plots to
# highlight genomic alterations across multiple samples.
#
# **Note**:
# Generating all copy number variation (CNV) plots is resource- and
# time-intensive. The process can take a significant amount of time, depending
# on the computational resources available.

# %%
analysis.precompute_cnvs()
cn_summary_path = calculate_cn_summary(analysis, "Diagnosis")

# %%
IPImage(filename=cn_summary_path)


# %% [markdown]
# -----------------------------------------------------------------------------
#
# ## Appendix
#
# The GEO annotation files were downloaded and manually curated. Below is the
# code used to download the GEO datasets and save their associated metadata:

# %% language="bash"
# pip install geoparse


# %%
import GEOparse

geo_numbers = datasets["scc"]["geo_ids"] + datasets["scc_test"]["geo_ids"]

raw_dir = geo_metadata_dir / "raw"
os.makedirs(raw_dir, exist_ok=True)

# Download and save metadata for each GEO series
for geo_nr in geo_numbers:
    gse = GEOparse.get_GEO(geo=geo_nr, destdir=raw_dir, how="brief")
    metadata = gse.phenotype_data
    print(metadata.head())
    metadata.to_csv(raw_dir / f"{geo_nr}_raw_metadata.csv", index=True)



"""Used to tests the performance of the mepylome package.

This script performs extraction of methylation information from all IDAT files
located in ~/mepylome/tests. It requires a preprocessing method as a
command-line argument (e.g., "illumina", "noob", "raw", "swan").

If '--save' is added it saves the extracted methylated and unmethylated
information to disk, enabling further comparison tests with the minfi package
using the ./test_idat_extraction.R script and validating correctness with the
./correctness.py script.

Usage:
    python test_idat_extraction.py illumina --save

    # To measure time and memory consumption
    /usr/bin/time -v python test_idat_extraction.py noob
"""

import sys
import time
from pathlib import Path

time0 = time.time()

from mepylome import MethylData, idat_basepaths
from mepylome.utils import ensure_directory_exists

time1 = time.time()


prep = None if len(sys.argv) < 2 else sys.argv[1]
preps = ["illumina", "noob", "raw", "swan"]
if prep not in preps:
    print(f"First command line argument must be in {preps}")
    print(f"Received: {prep}")
    sys.exit()


print(f"Time for importing mepylome: {time1 - time0} s")

HOME_DIR = Path.home()
TEST_DIR = Path(HOME_DIR, "mepylome", "tests")
TEST_OUTPUT_DIR = Path(HOME_DIR, "mepylome", "output_tests")

ensure_directory_exists(TEST_DIR)
ensure_directory_exists(TEST_OUTPUT_DIR)

idat_files = sorted(idat_basepaths(TEST_DIR))


time0 = time.time()

for idat_file in idat_files:
    methyl_data = MethylData(file=idat_file, prep=prep)

time1 = time.time()

N = len(idat_files)
tpc = (time1 - time0) / N

print(f"Time for analysis ({prep}): {time1 - time0} s ({N} cases)")
print(f"    Time per case: {tpc} s")

if len(sys.argv) < 3 or sys.argv[2] != "--save":
    print(
        "Exit script.\nIf you want to save output to disk for comparison "
        "with minfi, rerun this script by adding '--save'"
    )
    sys.exit()

print("\nSave results to disk.")

for idat_file in idat_files:
    probe = idat_file.name
    prefix_py = f"{probe}-mepylome-{prep}"
    prefix_r = f"{probe}-minfi-{prep}"

    methyl_path_py = Path(TEST_OUTPUT_DIR, f"{prefix_py}-methylated.csv")
    unmethyl_path_py = Path(TEST_OUTPUT_DIR, f"{prefix_py}-unmethylated.csv")

    methyl_data = MethylData(file=idat_file, prep=prep)
    methyl_data.methylated.sort_index().to_csv(methyl_path_py)
    methyl_data.unmethylated.sort_index().to_csv(unmethyl_path_py)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
#
# <img alt="Mepylome Logo" src="https://raw.githubusercontent.com/brj0/mepylome/main/mepylome/data/assets/mepylome.svg" width="300">
#
# Mepylome Tutorial
# =================
#
# ### Run This Notebook in Google Colab
#
# You can quickly open and run this notebook in Google Colab without any setup
# by clicking the link below.
#
# **Note**: The graphical user interface (GUI) features are limited in Google
# Colab.
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/brj0/mepylome/blob/main/examples/tutorial.ipynb)
#
#
# This notebook was automatically generated from the corresponding py-file
# with:
#
# ```bash
# jupytext --to ipynb tutorial.py
# ```

# %% [markdown]
# -----------------------------------------------------------------------------
# ### Install Required Packages
#
# To run the tutorial, install the following Python packages:
# - `mepylome` - the main toolkit for DNA-methylation analysis.
# - `ruptures` - used for segmentation calculations in CNV plots.
#
#
# Install these packages (may take 1 to 2 minutes) using the command below:

# %% language="bash"
#
# pip install -q mepylome ruptures


# %% [markdown]
# ### Run Tutorial

# %%
from pathlib import Path

from mepylome.analysis import MethylAnalysis
from mepylome.utils import setup_tutorial_files

DIR = Path.home() / "mepylome" / "tutorial"
ANALYSIS_DIR = DIR / "tutorial_analysis"
REFERENCE_DIR = DIR / "tutorial_reference"

setup_tutorial_files(ANALYSIS_DIR, REFERENCE_DIR)

analysis = MethylAnalysis(
    analysis_dir=ANALYSIS_DIR,
    reference_dir=REFERENCE_DIR,
    do_seg=True,
)
analysis.make_umap()

# %% [markdown]
# ### Graphical User Interface

# %%
analysis.run_app(open_tab=True)
